/***********************************************************
*                       pC/SFS 1.12b                       *
*            a FREE - small Serial File System             *
*                                                          *
*           by Frank Goetze  -  www.embedded-os.de         *
************************************************************
* Permission to use, copy, modify, and distribute this     *
* software in source and binary forms and its              *
* documentation for any purpose and without fee is hereby  *
* granted, provided that the above authors notice appear   *
* in all copies and that both that authors notice and this *
* permission notice appear in supporting documentation.    *
*                                                          *
* THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND   *
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT    *
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY    *
* AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.     *
* IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT,  *
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR             *
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,    *
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF     *
* USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER *
* CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN        *
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING           *
* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE   *
* USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY *
* OF SUCH DAMAGE.                                          *
************************************************************
*                        pC_SFS.c                          *
*                      File System                         *
************************************************************
* Revision-Info:                                           *
*                                                          *
* 0.80a     first alpa-release                             *
* 0.81a     code for create & delete link can be disabled  *
* 0.90a     handling of file attributes completed and      *
*           code uses found config values from file system *
*           if compatible to base settings                 *
*           ( "SFS_BLOCK_SIZE" & "<= SFS_MEMORY_SIZE" &    *
*            "sizeof(SFS_ENTRY)" )                         *
* 0.95b     internal function SFSChange_Entry() harmonised *
*           and the use of it readjusted                   *
* 0.96a     SFS_GetEntry() added, but only the hash of the *
*           name can be returned instead the name itself - *
*           and optional LOCK/UNLOCK added in LLdriver to  *
*           protect serial communication against timeouts  *
* 1.01b     parallel memory and I2C memory support added.  *
*           Rename of LL-driver files to split between     *
*           SPI, I2C and parallel.                         *
* 1.02a     correction in internal SFSChange_Entry() for   *
*           incorrect SFS_ChangeDir("/") handling,         *
* 1.10c     correction in internal SFSMem_Alloc() &        *
*           SFSCreate_Entry() on creation of a DIR         *
* 1.12b     correction in link-handling (linke-line) and   *
*           optional extension SFS_RemoveDirTree(..) added *
*                                                          *
***********************************************************/

/*
************************************************************
*                    additional bindings                   *
*              (SFS itself is bind by the port)            *
************************************************************
*/
#include "../inc/SFS_cfg.h"
#include "../inc/SFS_string.h"
#include "../inc/SFS_hash.h"

#ifdef _pC_OS_
 extern OS_TCB* OSTCBCur;
#endif


/***************** Constants im Code **********************/

U08 OS_CONST OS_ROMDATA SFSRevision[] = {"pC/SFS 1.12b"};

#define SFS_BLOCKMAP_SIZE    (SFS_MEMORY_SIZE / SFS_BLOCK_SIZE / 8)

SFS_ENTRY OS_CONST OS_ROMDATA ROOT = {                     // default SFS root entry
    '/',                                                   // for hash calculation of "/"
    SFS_ENTRIES_PER_DIR,                                   // DIR + entries per DIR (on DIR r/w will ignored)
    (SFS_MEMORY_SIZE / SFS_BLOCK_SIZE),                    // blocks over all (used memory)
    SFS_BLOCK_SIZE                                         // block-size
};

/****************** Globals im RAM ************************/

static SFS_ENTRY            ROOT_entry;                    // copy of SFS root entry (for 3 elements of block info)
static SFS_USER    OS_HUGE *SFS_User[SFS_MAX_USER];        // user control
static SFS_MUTEX_TYPE       SFS_mutex;                     // the SFS-mutex


/**************** Forward Prototypes **********************/

U08      SFSCheckUser(U08 *id) KC51_REENTRANT;
U08      SFSChange_Entry(U08 OS_HUGE *Name, U08 mode) KC51_REENTRANT;
U08      SFSCreate_Entry(U08 OS_HUGE *Name, U08 user, SFS_ATTR Attr, SFS_LONG size) KC51_REENTRANT;
U08      SFSFileOpend(SFS_ENTRY OS_HUGE *EntryPTR) KC51_REENTRANT;

U08      SFSMem_Format(void) KC51_REENTRANT;
U08      SFSMem_Check(void) KC51_REENTRANT;
U08      SFSMem_Alloc(SFS_LONG blocks, SFS_LONG *block_start) KC51_REENTRANT;
U08      SFSMem_Free(SFS_LONG block_start, SFS_LONG blocks) KC51_REENTRANT;

/*
************************************************************
*                    INITIALIZE SFS-MUTEX                  *
************************************************************
*/
static U08 SFSMutex_Init(SFS_MUTEX_TYPE OS_HUGE *mutex) KC51_REENTRANT
{
#ifdef _pC_OS_
    return(OS_SemInit(mutex, 1));                          // use a semaphore on pC/OS to test scheduling during two task in SFS
#else
    *mutex = 0; return(SFS_NO_ERR);
#endif
}

/*
************************************************************
*                      GET SFS-MUTEX                       *
************************************************************
*/
static U08 SFSMutex_Get(SFS_MUTEX_TYPE OS_HUGE *mutex) KC51_REENTRANT
{
#ifdef _pC_OS_
    return(OS_SemPend(mutex, OS_SUSPEND));                 // use a semaphore on pC/OS to test scheduling during two task in SFS
#else
    if (*mutex) return(1);                                 // as mark of "in use"
    //@TODO: disable interrupts here !
    *mutex = 1;
    //@TODO: enable interrupts here !
    return(SFS_NO_ERR);
#endif
}

/*
************************************************************
*                      PUT SFS-MUTEX                       *
************************************************************
*/
static U08 SFSMutex_Put(SFS_MUTEX_TYPE OS_HUGE *mutex) KC51_REENTRANT
{
#ifdef _pC_OS_
    return(OS_SemPost(mutex));
#else
    *mutex = 0; return(SFS_NO_ERR);
#endif
}

/*
************************************************************
*                    INITIALIZE SFS                        *
************************************************************
*/
U08 SFS_Init(void) KC51_REENTRANT
{
    U08 returnOk, User;

    returnOk = SFSMutex_Init(&SFS_mutex);
    if(returnOk == SFS_NO_ERR) {
        returnOk = SFSPort_Init();
        if(returnOk == SFS_NO_ERR) {
            for(User = 0; User < SFS_MAX_USER; User++) {
                SFS_User[User] = (SFS_USER OS_HUGE *)NULL;
            }
            returnOk = SFSPort_Read(0, (U08 OS_HUGE *)&ROOT_entry, sizeof(ROOT_entry));
            if(returnOk == SFS_NO_ERR) {                   // check ROOT at the begin of memory
                if((0 != SFS_hash_compare((U08 OS_HUGE *)"/", 1, ROOT_entry.name_hash)) ||
                   (ROOT_entry.block_start > ROOT.block_start) || 
                   (ROOT_entry.bytes_used != ROOT.bytes_used)  || 
                   (SFS_NO_ERR != SFSMem_Check())) {
#if (SFS_AUTO_FORMAT != 0)
                    returnOk = SFS_Format();
                    if(returnOk == SFS_NO_ERR) {
                        returnOk = SFSPort_Read(0, (U08 OS_HUGE *)&ROOT_entry, sizeof(ROOT_entry));
                        if(returnOk == SFS_NO_ERR) {       // check ROOT at the begin of memory
                            if((0 != SFS_hash_compare((U08 OS_HUGE *)"/", 1, ROOT_entry.name_hash)) ||
                               (ROOT_entry.attr__block_cnt != SFS_ENTRIES_PER_DIR) ||
                               (ROOT_entry.block_start != (SFS_MEMORY_SIZE / SFS_BLOCK_SIZE)) || 
                               (ROOT_entry.bytes_used != SFS_BLOCK_SIZE))
                                return(SFS_FORMAT_ERR);
                        }
                    }
#else
                    return(SFS_FORMAT_ERR);
#endif
                }
            }
        }
    }
    return(returnOk);
}

/*
************************************************************
*                    GET REVISION                          *
************************************************************
*/
U08 OS_HUGE *SFS_GetRev(void) KC51_REENTRANT
{
    return((U08 OS_HUGE *)SFSRevision);
}

/*
************************************************************
*                       FORMAT DRIVE                       *
************************************************************
*/
U08 SFS_Format(void) KC51_REENTRANT
{
    U08      returnOk, j;
    SFS_LONG i;

    returnOk = SFSMutex_Get(&SFS_mutex);
    if(returnOk == SFS_NO_ERR) {
        for(i = 0; i < SFS_MAX_USER; i++) {
            if(SFS_User[i] != (SFS_USER OS_HUGE *)NULL) {
                for(j = 0; j < SFS_HANDLES; j++) {
                   if (SFS_User[i]->used[j].openEntry.name_hash != 0) {
                       SFSMutex_Put(&SFS_mutex);
                       return(SFS_FILE_OPEN);
                   }
                }
            }
        }
        ROOT_entry           = ROOT;
        ROOT_entry.name_hash = SFS_hash_calc((U08 OS_HUGE *)"/", 1);
        returnOk = SFSPort_Fill(0, 0, SFS_MEMORY_SIZE);    // 1. clean-up / invalidate hole area
        if(returnOk == SFS_NO_ERR) {
            returnOk = SFSMem_Format();                    // 2. write blocks-bitmap
            if(returnOk == SFS_NO_ERR) {
                returnOk = SFSPort_Write(0, (U08 OS_HUGE *)&ROOT_entry, sizeof(ROOT_entry));
            }                                              // 3. write "ROOT" entry last (checked element)
        }	
        if(returnOk == SFS_NO_ERR) {
            returnOk = SFSMutex_Put(&SFS_mutex);
        } else {
            SFSMutex_Put(&SFS_mutex);
        }
    }
    return(returnOk);
}

/*
************************************************************
*                GET USER BY ID (internal)                 *
************************************************************
*/
U08 SFSGetUserByID(void) KC51_REENTRANT
{
#if OS_CRITICAL_SR
    OS_STK_TYPE  OS_sr;
#endif
    U08 i, id;

    OS_ENTER_CRITICAL();
    id = OS_USER_ID;
    for(i = 0; i < SFS_MAX_USER; i++) {
        if((SFS_User[i] != (SFS_USER OS_HUGE *)NULL) && (SFS_User[i]->Id == id)) {
            OS_EXIT_CRITICAL();
            return(i);
        }
    }
    OS_EXIT_CRITICAL();
    return(0xFF);
}

/*
************************************************************
*                 GET FREE USER (internal)                 *
************************************************************
*/
U08 SFSGetFreeUser(U08 *user) KC51_REENTRANT
{
#if OS_CRITICAL_SR
    OS_STK_TYPE  OS_sr;
#endif
    U08 i;

    OS_ENTER_CRITICAL();
    i = SFSGetUserByID();
    if(i == 0xFF) {
        for(i = 0; i < SFS_MAX_USER; i++) {
            if(SFS_User[i] == (SFS_USER OS_HUGE *)NULL) {
                break;
            }
        }
        if(i < SFS_MAX_USER) {
            *user = i;
            OS_EXIT_CRITICAL();
            return(SFS_NO_ERR);
        }
        i = 0xFF;
    }
    if(i == 0xFF) {
        OS_EXIT_CRITICAL();
        return(SFS_USR_OVF);
    }
    OS_EXIT_CRITICAL();
    return(SFS_DBL_USER);
}

/*
************************************************************
*                 CHECK USER (internal)                    *
************************************************************
*/
U08 SFSCheckUser(U08 *user) KC51_REENTRANT
{
    *user = SFSGetUserByID();
    if(*user != 0xFF) {
        if(SFS_User[*user] != (SFS_USER OS_HUGE *)NULL) {
            SFS_User[*user]->ErrNo = SFSMutex_Get(&SFS_mutex);
            return(SFS_User[*user]->ErrNo);
        } else {
            *user = 0xFF;
        }
    }
    return(SFS_NO_USER);
}

/*
************************************************************
*                        BECOME USER                       *
************************************************************
*/
U08 SFS_BecomeUser(SFS_USER OS_HUGE *SFSUser) KC51_REENTRANT
{
#if OS_CRITICAL_SR
    OS_STK_TYPE  OS_sr;
#endif
    U08          i, user;

    OS_ENTER_CRITICAL();
    i = SFSGetFreeUser(&user);
    if(i != SFS_NO_ERR) {
        OS_EXIT_CRITICAL();
        return(i);
    }
    if(SFS_User[user] == (SFS_USER OS_HUGE *)NULL) {
        SFS_User[user]               = SFSUser;
        SFS_User[user]->Id           = OS_USER_ID;
        OS_EXIT_CRITICAL();
        SFS_User[user]->currentEntry = ROOT_entry;
        SFS_User[user]->h_entrybi    = 0;
#if (SFS_TempDIR != 0)
        SFS_User[user]->TempDir.name_hash = 0;
#endif
        SFS_User[user]->ErrNo        = SFS_NO_ERR;
        for (i = 0; i < SFS_HANDLES; i++) {
            SFS_User[user]->used[i].openEntry.name_hash = 0;
            SFS_User[user]->used[i].o_entryPtr          = 0;
            SFS_User[user]->used[i].openOffset          = 0;
            SFS_User[user]->used[i].openMode            = 0;
        }
        return(SFS_NO_ERR);
    } else {
        OS_EXIT_CRITICAL();
        return(SFS_DBL_USER);
    }
}

/*
************************************************************
*                        CLOSE USER                        *
************************************************************
*/
U08 SFS_CloseUser(void) KC51_REENTRANT
{
    U08 i, user;

    user = SFSGetUserByID();
    if(user == 0xFF) {
        return(SFS_NO_USER);
    }
    for(i = 0; i < SFS_HANDLES; i++) {
        if(SFS_User[user]->used[i].openEntry.name_hash != 0) {
#if (SFS_AUTO_CLOSE != 0)
            if(SFS_NO_ERR != SFS_CloseFile(i+1))
                return(SFS_FILE_OPEN);
#else
            return(SFS_FILE_OPEN);
#endif
        }
    }
    SFS_User[user] = (SFS_USER OS_HUGE *)NULL;
    return(SFS_NO_ERR);
}

/*
************************************************************
*     GIVE THE ERROR-CODE FROM OPEN / READ / WRITE / ..    *
************************************************************
*/
U08 SFS_GetErrNo(void) KC51_REENTRANT
{
    U08 returnOk, user;

    user = SFSGetUserByID();
    if(user == 0xFF) {
        return(SFS_NO_USER);
    }
    returnOk = SFS_User[user]->ErrNo;
    SFS_User[user]->ErrNo = SFS_NO_ERR;
    return(returnOk);
}

/*
************************************************************
*                         OPEN FILE                        *
************************************************************
*/
SFS_HANDLE SFS_OpenFile(U08 OS_HUGE *Name, U08 Mode) KC51_REENTRANT
{
    SFS_ENTRY   currDir;
    SFS_LONG    ptr;
    U08         returnOk, i, j, user;

    if(SFS_NO_ERR != SFSCheckUser(&user)) {
        return(0);                                         // no correct user (no ErrNo possible)
    }
    Mode &= SFS_ACC_RW;
    if(Mode == 0) {
        SFSMutex_Put(&SFS_mutex);
        SFS_User[user]->ErrNo = SFS_WRONG_A;
        return(0);
    }
    j = (U08)-1;
    currDir = SFS_User[user]->currentEntry;                // save actual position
    returnOk = SFSChange_Entry(Name, 1);                   // go into file / into linked file
    if(returnOk == SFS_NO_ERR) {                           // go into PATH
        i = ((U08)(SFS_User[user]->currentEntry.attr__block_cnt >> (8 * (sizeof(SFS_LONG) - 1))) & SFS_ATTR_RW);
        if((i & Mode) != 0) {
            for(i = 0; i < SFS_MAX_USER; i++) {
                if(SFS_User[i] != (SFS_USER OS_HUGE *)NULL) {
                    for(j = 0; j < SFS_HANDLES; j++) {
                        if(((SFS_User[i]->used[j].openEntry.name_hash != 0) &&
                           (SFS_User[i]->used[j].openEntry.block_start == SFS_User[user]->currentEntry.block_start)) &&
                          ((SFS_User[i]->used[j].openMode != SFS_ACC_RO) ||
                           (Mode != SFS_ACC_RO))) {
                            SFS_User[user]->currentEntry = currDir;  // restore actual position
                            SFS_User[user]->ErrNo = SFS_FILE_OPEN;
                            SFSMutex_Put(&SFS_mutex);
                            return(0);
                        }
                    }
                }
            }
            if(returnOk == SFS_NO_ERR) {
                i = 0; j = SFS_HANDLES;
                while(i < j) {
                    if((SFS_User[user]->used[i].openEntry.name_hash == 0) && (j == SFS_HANDLES)) {
                        j = i;
                    }
                    i++;
                }
                if(j != SFS_HANDLES) {
                    SFS_User[user]->used[j].openEntry  = SFS_User[user]->currentEntry;
                    ptr = ((SFS_User[user]->h_entrybi & (((SFS_LONG)-1) >> 8)) * ROOT_entry.bytes_used) + ((SFS_User[user]->h_entrybi >> (8 * (sizeof(SFS_LONG) - 1))) * sizeof(SFS_ENTRY));
                    SFS_User[user]->used[j].o_entryPtr = ptr;
                    SFS_User[user]->used[j].openMode   = Mode;
                    SFS_User[user]->used[j].openOffset = 0;
                } else {                                   // no free handle-field
                    returnOk = SFS_USR_OVF;
                }
            }
        } else {                                           // file is RO or WO
            if(Mode == SFS_ACC_RO) {
                returnOk = SFS_FILE_WO;
            } else {
                returnOk = SFS_FILE_RO;
            }
        }
    }
    SFS_User[user]->currentEntry = currDir;                // restore actual position
    SFS_User[user]->ErrNo = returnOk;
    SFSMutex_Put(&SFS_mutex);
    if(returnOk == SFS_NO_ERR) {
        return((SFS_HANDLE)(j+1));
    }
    return(0);
}

/*
************************************************************
*                         CLOSE FILE                       *
************************************************************
*/
U08 SFS_CloseFile(SFS_HANDLE handl) KC51_REENTRANT
{
    U08 returnOk, user;

    returnOk = SFSCheckUser(&user);
    if(returnOk != SFS_NO_ERR) {
        return(returnOk);
    }
    if((handl == 0) || (handl > SFS_HANDLES) ||
       (SFS_User[user]->used[handl-1].openEntry.name_hash == 0)) {
        SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
        return(SFS_NO_USER);                               // if handle invalid or no opened file
    }
    handl -= 1;
    if((SFS_User[user]->used[handl].openEntry.attr__block_cnt) & ((SFS_LONG)SFS_ATTR_FILE << (8 * (sizeof(SFS_LONG) - 1)))) {
        SFS_User[user]->used[handl].openEntry.name_hash = 0;
        SFS_User[user]->used[handl].openMode   = 0;
        SFS_User[user]->used[handl].openOffset = 0;
        returnOk = SFS_NO_ERR;
    } else {
        returnOk = SFS_NO_FILE;                            // behind this handle is no file
    }
    SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                         RENAME FILE                      *
************************************************************
*/
U08 SFS_RenameFile(U08 OS_HUGE *OldName, U08 OS_HUGE *NewName) KC51_REENTRANT
{
    SFS_ENTRY   currDir, nEntry;
    SFS_LONG    Pos, o_bi, ptr;
    U08         returnOk, user;

    returnOk = SFSCheckUser(&user);
    if(returnOk != SFS_NO_ERR) {
        return(returnOk);
    }
    currDir = SFS_User[user]->currentEntry;                // save actual position and go into PATH
    returnOk = SFSChange_Entry(OldName, 1);                // go into file / into linked file
    if(returnOk == SFS_NO_ERR) {                           // if valid old name
        o_bi = SFS_User[user]->h_entrybi;                  // save location of entry in dir block
        Pos = SFSstrrpos(NewName,'/');                     // search last '/' in PATH or full name and check
        if(Pos == (SFS_LONG)(0-1)) {
            Pos = 0;
        } else {
            Pos++;                                         // update name if NewName not exist
        }
        if(Pos < SFSstrlen(NewName)) {
            returnOk = SFSChange_Entry(&NewName[Pos], 1);  // go into file / into linked file
            if(returnOk == SFS_NOT_EXIST) {                // not found
                nEntry.name_hash = SFS_hash_calc(&NewName[Pos], SFSstrlen(&NewName[Pos]));
                ptr = ((o_bi & (((SFS_LONG)-1) >> 8)) * ROOT_entry.bytes_used) + ((o_bi >> (8 * (sizeof(SFS_LONG) - 1))) * sizeof(SFS_ENTRY));
                returnOk = SFSPort_Write(ptr, (U08 OS_HUGE *)&nEntry, sizeof(nEntry.name_hash));
            } else {
                returnOk = SFS_NAME_EXIST;                 // new-name exist
            }
        } else {
            returnOk = SFS_PATH_ERR;                       // error in new-name
        }
    } else {
        returnOk = SFS_PATH_ERR;                           // no existing PATH given
    }
    SFS_User[user]->currentEntry = currDir;                // restore actual position
    SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                    CHANGE FILE-ATTRIBS                   *
************************************************************
*/
U08 SFS_AttribFile(U08 OS_HUGE *Name, SFS_ATTR Attribs) KC51_REENTRANT
{
    SFS_ENTRY   currDir;
    SFS_LONG    b;
    SFS_ATTR    a;
    U08         returnOk, user;

    returnOk = SFSCheckUser(&user);
    if(returnOk != SFS_NO_ERR) {
        return(returnOk);
    }
    Attribs &= SFS_ATTR_RW;
    if(Attribs == 0) {
        SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
        return(SFS_WRONG_A);
    }
    currDir = SFS_User[user]->currentEntry;                // save actual position and go into PATH
    returnOk = SFSChange_Entry(Name, 1);                   // go into file / into linked file
    if(returnOk == SFS_NO_ERR) {                           // if valid path and filename - go into
        returnOk = SFSFileOpend(&SFS_User[user]->currentEntry);
        if(returnOk == SFS_NO_ERR) {
            a  = ((SFS_User[user]->currentEntry.attr__block_cnt >> (8 * (sizeof(SFS_LONG) - 1))) & (SFS_ATTR_LINK | SFS_ATTR_LINKS_M)) | (Attribs & SFS_ATTR_RW); 
            a |= SFS_ATTR_FILE;                            // update attributes and write-back
            SFS_User[user]->currentEntry.attr__block_cnt = (((SFS_LONG)a) << (8 * (sizeof(SFS_LONG) - 1))) | (SFS_User[user]->currentEntry.attr__block_cnt & (((SFS_LONG)(-1)) >> 8));
            b = ((SFS_User[user]->h_entrybi & (((SFS_LONG)-1) >> 8)) * ROOT_entry.bytes_used) + ((SFS_User[user]->h_entrybi >> (8 * (sizeof(SFS_LONG) - 1))) * sizeof(SFS_ENTRY));
            returnOk = SFSPort_Write((U32)&((SFS_ENTRY *)b)->attr__block_cnt, (U08 OS_HUGE *)&SFS_User[user]->currentEntry.attr__block_cnt, sizeof(SFS_LONG));
        }
    }
    SFS_User[user]->currentEntry = currDir;
    SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                      GET FILE-ATTRIBS                    *
************************************************************
*/
SFS_ATTR SFS_GetFileAttrib(U08 OS_HUGE *Name) KC51_REENTRANT
{
    SFS_ENTRY    currDir;
    SFS_ATTR     a;
    U08          returnOk, user;

    if(SFS_NO_ERR != SFSCheckUser(&user)) {
        return(0);                                         // no correct user
    }
    currDir = SFS_User[user]->currentEntry;                // save actual position and go into PATH
    returnOk = SFSChange_Entry(Name, 1);                   // go into file / into linked file
    if(returnOk == SFS_NO_ERR) {
        returnOk = SFSFileOpend(&SFS_User[user]->currentEntry);
        if(returnOk == SFS_NO_ERR) {
            a = (SFS_User[user]->currentEntry.attr__block_cnt >> (8 * (sizeof(SFS_LONG) - 1))) & SFS_ATTR_RW;
            SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
            return(a);
        }
    }
    SFS_User[user]->ErrNo = returnOk;
    SFS_User[user]->currentEntry = currDir;
    SFSMutex_Put(&SFS_mutex);
    return((SFS_ATTR)(-1));
}

/*
************************************************************
*                GET CURRENT SIZE OF FILE                  *
************************************************************
*/
SFS_LONG SFS_GetFileSize(U08 OS_HUGE *Name) KC51_REENTRANT
{
    SFS_ENTRY    currDir, FileEntry;
    SFS_LONG     len;
    U08          returnOk, user;

    if(SFS_NO_ERR != SFSCheckUser(&user)) {
        return(0);
    }
    currDir = SFS_User[user]->currentEntry;                // save actual position and go into PATH
    returnOk = SFSChange_Entry(Name, 1);                   // go into file / into linked file
    if(returnOk == SFS_NO_ERR) {                           // if valid name - go into
        returnOk = SFSFileOpend(&SFS_User[user]->currentEntry);
    }
    FileEntry = SFS_User[user]->currentEntry;
    SFS_User[user]->currentEntry = currDir;                // restore saved position
    SFS_User[user]->ErrNo = returnOk;
    if(returnOk != SFS_NO_ERR) {
        SFSMutex_Put(&SFS_mutex);
        return(0);                                         // error
    }
    len = FileEntry.bytes_used;
    SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
    return(len);
}

/*
************************************************************
*              GET ALLOCATED SIZE OF FILE                  *
************************************************************
*/
SFS_LONG SFS_GetMaxFileSize(U08 OS_HUGE *Name) KC51_REENTRANT
{
    SFS_ENTRY    currDir, FileEntry;
    SFS_LONG     len;
    U08          returnOk, user;

    if(SFS_NO_ERR != SFSCheckUser(&user)) {
        return(0);
    }
    currDir = SFS_User[user]->currentEntry;                // save actual position and go into PATH
    returnOk = SFSChange_Entry(Name, 1);                   // go into file / into linked file
    if(returnOk == SFS_NO_ERR) {                           // if valid name - go into
        returnOk = SFSFileOpend(&SFS_User[user]->currentEntry);
    }
    FileEntry = SFS_User[user]->currentEntry;
    SFS_User[user]->currentEntry = currDir;                // restore saved position
    SFS_User[user]->ErrNo = returnOk;
    if(returnOk != SFS_NO_ERR) {
        SFSMutex_Put(&SFS_mutex);
        return(0);                                         // error
    }
    len = (FileEntry.attr__block_cnt & (((SFS_LONG)-1) >> 8)) * ROOT_entry.bytes_used;
    SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
    return(len);
}

/*
************************************************************
*               SET ACTUAL PTR IN OPEN FILE                *
************************************************************
*/
U08 SFS_SeekFile(SFS_LONG offset, SFS_HANDLE handl) KC51_REENTRANT
{
    SFS_LONG             Doffset;
    U08                  returnOk, user;

    returnOk = SFSCheckUser(&user);
    if(returnOk != SFS_NO_ERR) {
        return(returnOk);
    }
    if((handl == 0) || (handl > SFS_HANDLES) ||
       (SFS_User[user]->used[handl-1].openEntry.name_hash == 0)) {
        return(SFS_NO_USER);                               // check if file opened
    }
    if(SFS_User[user]->used[handl-1].openEntry.attr__block_cnt & ((SFS_LONG)SFS_ATTR_FILE << (8 * (sizeof(SFS_LONG) - 1)))) {
            if(SFS_User[user]->used[handl-1].openEntry.block_start == 0) {
            SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
            return(SFS_NO_DATA);
        }
        if(offset != SFS_SEEK_EOF) {                       // calculate offset
            Doffset = offset;
        } else {
            Doffset = SFS_User[user]->used[handl-1].openEntry.bytes_used;
        }
        if(Doffset > (SFS_User[user]->used[handl-1].openEntry.attr__block_cnt & (((SFS_LONG)-1) >> 8)) * ROOT_entry.bytes_used) {
            SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
            return(SFS_FILE_EOF);                          // offset is larger allocated-size
        }
        SFS_User[user]->used[handl-1].openOffset = Doffset;
    } else {
        returnOk = SFS_NO_FILE;                            // no file opened
    }
    SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
    return(returnOk);
}

/*
************************************************************
*            SET EOF IN OPEN FILE ON ACTUAL PTR            *
************************************************************
*/
U08 SFS_SetEOF(SFS_HANDLE handl) KC51_REENTRANT
{
    U08         returnOk, user;

    returnOk = SFSCheckUser(&user);
    if(returnOk != SFS_NO_ERR) {
        return(returnOk);
    }
    if((handl == 0) || (handl > SFS_HANDLES) ||
       (SFS_User[user]->used[handl-1].openEntry.name_hash == 0)) {
        returnOk = SFS_NO_USER;
    } else {                                               // check if file opened
        if(SFS_User[user]->used[handl-1].openEntry.attr__block_cnt & ((SFS_LONG)SFS_ATTR_FILE << (8 * (sizeof(SFS_LONG) - 1)))) {
            if(SFS_User[user]->used[handl-1].openEntry.block_start == 0) {
                returnOk = SFS_NO_DATA;
            }
            if(!(SFS_User[user]->used[handl-1].openEntry.attr__block_cnt & ((SFS_LONG)SFS_ATTR_WO << (8 * (sizeof(SFS_LONG) - 1)))) || (SFS_User[user]->used[handl-1].openMode == SFS_ACC_RO)) {
                returnOk = SFS_FILE_RO;                    // check if not RO-File or opened in RO-mode
            }
        } else {
            returnOk = SFS_NO_FILE;                        // no file opened
        }
    }
    if(returnOk == SFS_NO_ERR) {
        SFS_User[user]->used[handl-1].openEntry.bytes_used = SFS_User[user]->used[handl-1].openOffset;
        returnOk = SFSPort_Write(SFS_User[user]->used[handl-1].o_entryPtr, (U08 OS_HUGE *)(&SFS_User[user]->used[handl-1].openEntry), sizeof(SFS_ENTRY));
    }
    SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex); 
    return(returnOk);
}

/*
************************************************************
*               GET ACTUAL PTR IN OPEN FILE                *
************************************************************
*/
SFS_LONG SFS_TellFile(SFS_HANDLE handl) KC51_REENTRANT
{
    U08         returnOk, user;
    SFS_LONG    pos;

    returnOk = SFSCheckUser(&user);
    if(returnOk == SFS_NO_ERR) {
        if((handl == 0) || (handl > SFS_HANDLES) ||
           (SFS_User[user]->used[handl-1].openEntry.name_hash == 0)) {
            returnOk = SFS_NO_USER;
        } else {                                           // check if file opened
            if(SFS_User[user]->used[handl-1].openEntry.attr__block_cnt & ((SFS_LONG)SFS_ATTR_FILE << (8 * (sizeof(SFS_LONG) - 1)))) {
                if(SFS_User[user]->used[handl-1].openEntry.block_start == 0) {
                    returnOk = SFS_NO_DATA;
                }
            } else {
                returnOk = SFS_NO_FILE;                    // no file opened
            }
        }
    }
    SFS_User[user]->ErrNo = returnOk;
    if(returnOk == SFS_NO_ERR) {
        pos = SFS_User[user]->used[handl-1].openOffset;
        SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
        return(pos);
    }
    SFSMutex_Put(&SFS_mutex);
    return(0);
}

/*
************************************************************
*                         READ FILE                        *
************************************************************
*/
SFS_LONG SFS_ReadFile(U08 OS_HUGE *dest, SFS_LONG size, SFS_HANDLE handl) KC51_REENTRANT
{
    U08         user, attr;
    SFS_LONG    offset, rsize;

    if(SFS_NO_ERR != SFSCheckUser(&user)) {
        return(0);
    }
    if((handl == 0) || (handl > SFS_HANDLES) ||
       (SFS_User[user]->used[handl-1].openEntry.name_hash == 0)) {
        SFS_User[user]->ErrNo = SFS_NO_USER;
        SFSMutex_Put(&SFS_mutex);
        return(0);                                         // check if file opened
    }
    attr = SFS_User[user]->used[handl-1].openEntry.attr__block_cnt >> (8 * (sizeof(SFS_LONG) - 1));
    if(attr & SFS_ATTR_FILE) {
        if(!(attr & SFS_ATTR_RO) || (SFS_User[user]->used[handl-1].openMode == SFS_ACC_WO)) {
            SFS_User[user]->ErrNo = SFS_FILE_WO;
            SFSMutex_Put(&SFS_mutex);
            return(0);                                     // check file to read-access
        }
        if(SFS_User[user]->used[handl-1].openEntry.block_start == 0) {
            SFS_User[user]->ErrNo = SFS_NO_DATA;
            SFSMutex_Put(&SFS_mutex);
            return(0);                                     // check if file have a Data-Block
        }
        offset = SFS_User[user]->used[handl-1].openOffset;
        rsize = ((offset + size) > SFS_User[user]->used[handl-1].openEntry.bytes_used)? SFS_User[user]->used[handl-1].openEntry.bytes_used - offset : size;
        SFS_User[user]->ErrNo = SFSPort_Read((SFS_User[user]->used[handl-1].openEntry.block_start * ROOT_entry.bytes_used) + offset, dest, rsize);
        SFS_User[user]->used[handl-1].openOffset += rsize; // set actual offset
    } else {
        SFS_User[user]->ErrNo = SFS_NO_FILE;
        SFSMutex_Put(&SFS_mutex);
        return(0);                                         // no file opened
    }
    SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
    return(rsize);
}

/*
************************************************************
*                         WRITE FILE                       *
************************************************************
*/
SFS_LONG SFS_WriteFile(U08 OS_HUGE *src, SFS_LONG size, SFS_HANDLE handl) KC51_REENTRANT
{
    U08         user, attr;
    SFS_LONG    offset;

    if(SFS_NO_ERR != SFSCheckUser(&user)) {
        return(0);
    }
    if((handl == 0) || (handl > SFS_HANDLES) ||
       (SFS_User[user]->used[handl-1].openEntry.name_hash == 0)) {
        SFS_User[user]->ErrNo = SFS_NO_USER;
        SFSMutex_Put(&SFS_mutex);
        return(0);                                         // check if file opened
    }
    attr = SFS_User[user]->used[handl-1].openEntry.attr__block_cnt >> (8 * (sizeof(SFS_LONG) - 1));
    if(attr & SFS_ATTR_FILE) {
        if(!(attr & SFS_ATTR_WO) || (SFS_User[user]->used[handl-1].openMode == SFS_ACC_RO)) {
            SFS_User[user]->ErrNo = SFS_FILE_RO;
            SFSMutex_Put(&SFS_mutex);
            return(0);                                     // check file to write-access
        }
        if(SFS_User[user]->used[handl-1].openEntry.block_start == 0) {
            SFS_User[user]->ErrNo = SFS_NO_DATA;
            SFSMutex_Put(&SFS_mutex);
            return(0);                                     // check if file have a Data-Block
        }
        offset = SFS_User[user]->used[handl-1].openOffset;
        if((offset + size) > ((SFS_User[user]->used[handl-1].openEntry.attr__block_cnt & (((SFS_LONG)-1) >> 8)) * ROOT_entry.bytes_used)) {
            SFS_User[user]->ErrNo = SFS_FILE_EOF;
            SFSMutex_Put(&SFS_mutex);
            return(0);                                     // offset + size is larger allocated-size
        }
        SFS_User[user]->ErrNo = SFSPort_Write((SFS_User[user]->used[handl-1].openEntry.block_start * ROOT_entry.bytes_used) + offset, src, size);
        offset += size;
        SFS_User[user]->used[handl-1].openOffset = offset; // set actual offset
        if(SFS_User[user]->used[handl-1].openEntry.bytes_used < offset) {
            SFS_User[user]->used[handl-1].openEntry.bytes_used = offset;
            SFS_User[user]->ErrNo = SFSPort_Write(SFS_User[user]->used[handl-1].o_entryPtr, (U08 OS_HUGE *)(&SFS_User[user]->used[handl-1].openEntry), sizeof(SFS_ENTRY));
            if(SFS_User[user]->ErrNo != SFS_NO_ERR) {
                SFSMutex_Put(&SFS_mutex);
                return(size);
            }
        }
    } else {
        SFS_User[user]->ErrNo = SFS_NO_FILE;
        SFSMutex_Put(&SFS_mutex);
        return(0);                                         // no file opened
    }
    SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
    return(size);
}

#if (SFS_LINKS != 0)
/*
************************************************************
*                         CREATE LINK                      *
************************************************************
*/
U08 SFS_CreateLink(U08 OS_HUGE *OrgName, U08 OS_HUGE *Name) KC51_REENTRANT
{
    SFS_ENTRY   currDir, orgEntry, linkEntry;
    SFS_LONG    orgEntrybi, b;
    U08         returnOk, user, i, a;
    SFS_LONG    Pos = 0;

    returnOk = SFSCheckUser(&user);
    if(returnOk != SFS_NO_ERR) {
        return(returnOk);
    }
    currDir = SFS_User[user]->currentEntry;                // save actual position and go into PATH
    returnOk = SFSChange_Entry(OrgName, 4);                // go into * / stay on linked *
    if(returnOk == SFS_NO_ERR) {                           // create Entry
        orgEntry   = SFS_User[user]->currentEntry;
        orgEntrybi = SFS_User[user]->h_entrybi;
        SFS_User[user]->currentEntry = currDir;            // restore saved position
        returnOk = SFSChange_Entry(Name, 0);               // go up to last part of path
        if(returnOk == SFS_NO_ERR) {
            Pos = SFSstrrpos(Name, '/');
            if(Pos == (SFS_LONG)(0-1)) {
                Pos = 0;
            } else {
                Pos++;
            }
            if(Pos < SFSstrlen(Name)) {                    // update link-cnt & create Entry
                i = (orgEntry.attr__block_cnt >> (8 * (sizeof(SFS_LONG) - 1))) & SFS_ATTR_LINKS_M;
                if(i < 15) {
                    i++;
                    orgEntry.attr__block_cnt = (orgEntry.attr__block_cnt & ~(SFS_ATTR_LINKS_M << (8 * (sizeof(SFS_LONG) - 1)))) | (i << (8 * (sizeof(SFS_LONG) - 1)));
                    b = ((orgEntrybi & (((SFS_LONG)-1) >> 8)) * ROOT_entry.bytes_used) + ((orgEntrybi >> (8 * (sizeof(SFS_LONG) - 1))) * sizeof(SFS_ENTRY));
                    returnOk = SFSPort_Write(b, (U08 OS_HUGE *)&orgEntry, sizeof(SFS_ENTRY));
                    if(returnOk == SFS_NO_ERR) {
                        a = SFS_ATTR_LINK | ((orgEntry.attr__block_cnt >> (8 * (sizeof(SFS_LONG) - 1))) & (SFS_ATTR_DIR | SFS_ATTR_FILE));
                        linkEntry.name_hash       = SFS_hash_calc(&Name[Pos], SFSstrlen(Name) - Pos);
                        linkEntry.attr__block_cnt = a << (8 * (sizeof(SFS_LONG) - 1));
                        linkEntry.block_start     = orgEntrybi;
                        linkEntry.bytes_used      = 0;
                        b = ((SFS_User[user]->h_entrybi & (((SFS_LONG)-1) >> 8)) * ROOT_entry.bytes_used) + ((SFS_User[user]->h_entrybi >> (8 * (sizeof(SFS_LONG) - 1))) * sizeof(SFS_ENTRY));
                        returnOk = SFSPort_Write(b, (U08 OS_HUGE *)&linkEntry, sizeof(SFS_ENTRY));
                    }
                } else {
                    returnOk = SFS_MAX_LINK;               // max link count
                }
            } else {
                returnOk = SFS_PATH_ERR;                   // no valid link-name given
            }
        }
    }
    SFS_User[user]->currentEntry = currDir;
    SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                         REMOVE LINK                      *
************************************************************
*/
U08 SFS_RemoveLink(U08 OS_HUGE *Name) KC51_REENTRANT
{
    SFS_ENTRY   currDir, orgEntry, linkEntry;
    SFS_LONG    b;
    U08         returnOk, i, user;

    returnOk = SFSCheckUser(&user);
    if(returnOk != SFS_NO_ERR) {
        return(returnOk);
    }
    currDir = SFS_User[user]->currentEntry;                // save actual position and go to link (not into link)
    returnOk = SFSChange_Entry(Name, 4);                   // go into * / stay on link-*
    if(returnOk == SFS_NO_ERR) {
        linkEntry   = SFS_User[user]->currentEntry;
        if(((linkEntry.attr__block_cnt >> (8 * (sizeof(SFS_LONG) - 1))) & SFS_ATTR_LINKS_M) == 0) {
            if(linkEntry.block_start != 0) {
                b = ((linkEntry.block_start & (((SFS_LONG)-1) >> 8)) * ROOT_entry.bytes_used) + ((linkEntry.block_start >> (8 * (sizeof(SFS_LONG) - 1))) * sizeof(SFS_ENTRY));
                returnOk = SFSPort_Read(b, (U08 OS_HUGE *)&orgEntry, sizeof(SFS_ENTRY));
                if(returnOk == SFS_NO_ERR) {
                    i = (orgEntry.attr__block_cnt >> (8 * (sizeof(SFS_LONG) - 1))) & SFS_ATTR_LINKS_M;
                    if(i > 0) {
                        i--;
                        orgEntry.attr__block_cnt = (orgEntry.attr__block_cnt & ~(SFS_ATTR_LINKS_M << (8 * (sizeof(SFS_LONG) - 1)))) | (i << (8 * (sizeof(SFS_LONG) - 1)));
                        returnOk = SFSPort_Write(b, (U08 OS_HUGE *)&orgEntry, sizeof(SFS_ENTRY));
                        if(returnOk == SFS_NO_ERR) {
                            linkEntry.name_hash = 0;
                            b = ((SFS_User[user]->h_entrybi & (((SFS_LONG)-1) >> 8)) * ROOT_entry.bytes_used) + ((SFS_User[user]->h_entrybi >> (8 * (sizeof(SFS_LONG) - 1))) * sizeof(SFS_ENTRY));
                            returnOk = SFSPort_Write(b, (U08 OS_HUGE *)&linkEntry, sizeof(SFS_ENTRY));
                        }
                    } else {
                        returnOk = SFS_LINK_ERR;
                    }
                }
            } else {
                returnOk = SFS_NO_LINK;                    // no link-PTR in Link-Entry
            }
        } else {
            returnOk = SFS_LINKED;                         // link is linked
        }
    }
    SFS_User[user]->currentEntry = currDir;                // restore saved position
    SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
    return(returnOk);
}
#endif

/*
************************************************************
*                         CREATE DIR                       *
************************************************************
*/
U08 SFS_CreateDir(U08 OS_HUGE *Name) KC51_REENTRANT
{
    SFS_ENTRY   currDir;
    U08         returnOk, user;
    SFS_LONG    Pos = 0;

    returnOk = SFSCheckUser(&user);
    if(returnOk != SFS_NO_ERR) {
        return(returnOk);
    }
    currDir = SFS_User[user]->currentEntry;                // save actual position and go into PATH
    returnOk = SFSChange_Entry(Name, 0);                   // go up to last part of path
    if(returnOk == SFS_NO_ERR) {                           // create Entry
        Pos = SFSstrrpos(Name, '/');
        if(Pos == (SFS_LONG)(0-1)) {
            Pos = 0;
        } else {
            Pos++;
        }
        if(Pos < SFSstrlen(Name)) {                        // create Entry
            returnOk = SFSCreate_Entry(&Name[Pos], user, SFS_ATTR_DIR, ROOT_entry.attr__block_cnt * sizeof(SFS_ENTRY));
        } else {
            returnOk = SFS_PATH_ERR;
        }
    }
    SFS_User[user]->currentEntry = currDir;
    SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                         REMOVE DIR                       *
************************************************************
*/
U08 SFS_RemoveDir(U08 OS_HUGE *Name) KC51_REENTRANT
{
    SFS_ENTRY   currDir, checkEntry;
    SFS_LONG    i, ptr;
    U08         returnOk, user;

    returnOk = SFSCheckUser(&user);
    if(returnOk != SFS_NO_ERR) {
        return(returnOk);
    }
    currDir = SFS_User[user]->currentEntry;                // save actual position and go into PATH
    returnOk = SFSChange_Entry(Name, 2);                   // go into dir / into linked dir
    if(returnOk == SFS_NO_ERR) {                           // Delete Entry if no sub-elements and not the ROOT
        if((SFS_User[user]->currentEntry.attr__block_cnt & (SFS_ATTR_LINKS_M << (8 * (sizeof(SFS_LONG) - 1)))) == 0) {
            i = 1;                                         // check that no entry is in this dir (expect "..")
            while(i < ROOT_entry.attr__block_cnt) {
                returnOk = SFSPort_Read((SFS_User[user]->currentEntry.block_start * ROOT_entry.bytes_used) + (i * sizeof(SFS_ENTRY)), (U08 OS_HUGE *)&checkEntry, sizeof(SFS_ENTRY));
                if(returnOk == SFS_NO_ERR) {
                    if(checkEntry.name_hash != 0) break;
                } else {
                    SFS_User[user]->currentEntry = currDir; // restore saved position
                    SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
                    return(returnOk);
                }
                i++;
            }
            if(i >= ROOT_entry.attr__block_cnt) {
                for(i = 0; i < SFS_MAX_USER; i++) {        // check all user with their currentDir / tempDir
                    if(SFS_User[i] != (SFS_USER OS_HUGE *)NULL) {
                        if(i != user) {
                            if(SFS_User[i]->currentEntry.block_start == SFS_User[user]->currentEntry.block_start) break;
#if (SFS_TempDIR != 0)
                            if(SFS_User[i]->TempDir.block_start == SFS_User[user]->TempDir.block_start) break;
#endif
                        } else {
                            if(currDir.block_start == SFS_User[user]->currentEntry.block_start) break;
                        }
                    }
                }
                if(i >= SFS_MAX_USER) {                    // delete the dir entry blocks and the entry itself
                    ptr = ((ROOT_entry.attr__block_cnt * sizeof(SFS_ENTRY)) / ROOT_entry.bytes_used) + (((ROOT_entry.attr__block_cnt * sizeof(SFS_ENTRY)) % ROOT_entry.bytes_used)? 1 : 0);
                    returnOk = SFSMem_Free(SFS_User[user]->currentEntry.block_start, ptr);
                    if(returnOk == SFS_NO_ERR) {
                        checkEntry.name_hash     = 0;
                        ptr = ((SFS_User[user]->h_entrybi & (((SFS_LONG)-1) >> 8)) * ROOT_entry.bytes_used) + ((SFS_User[user]->h_entrybi >> (8 * (sizeof(SFS_LONG) - 1))) * sizeof(SFS_ENTRY));
                        returnOk = SFSPort_Write(ptr, (U08 OS_HUGE *)&checkEntry, sizeof(checkEntry.name_hash));
                    }
                } else {
                    returnOk = SFS_TMP_DIR;                // dir is in use (current/temp by a user)
                }
            } else {
                returnOk = SFS_NOT_EMPTY;                  // dir is not empty
            }
        } else {
            returnOk = SFS_LINKED;                         // entry is linked
        }
    }
    SFS_User[user]->currentEntry = currDir;                // restore saved position
    SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                         RENAME DIR                       *
************************************************************
*/
U08 SFS_RenameDir(U08 OS_HUGE *OldName, U08 OS_HUGE *NewName) KC51_REENTRANT
{
    SFS_ENTRY   currDir, nEntry;
    SFS_LONG    Pos, o_bi, ptr;
    U08         returnOk, user;

    returnOk = SFSCheckUser(&user);
    if(returnOk != SFS_NO_ERR) {
        return(returnOk);
    }
    currDir = SFS_User[user]->currentEntry;                // save actual position and go into PATH
    returnOk = SFSChange_Entry(OldName, 2);                // go into dir / into linked dir
    if(returnOk == SFS_NO_ERR) {                           // if valid old name
        o_bi = SFS_User[user]->h_entrybi;                  // save location of entry in dir block
        Pos = SFSstrrpos(NewName,'/');                     // search last '/' in PATH or full name and check
        if(Pos == (SFS_LONG)(0-1)) {
            Pos = 0;
        } else {
            Pos++;                                         // update name if NewName not exist
        }
        if(Pos < SFSstrlen(NewName)) {                     // search NewName in prev dir
            returnOk = SFSChange_Entry((U08 OS_HUGE *)"..", 3);
            if(returnOk == SFS_NO_ERR) {                   // go into * / into link-*
                returnOk = SFSChange_Entry(&NewName[Pos], 3);
                if(returnOk == SFS_NOT_EXIST) {            // not found ...
                    nEntry.name_hash = SFS_hash_calc(&NewName[Pos], SFSstrlen(&NewName[Pos]));
                    ptr = ((o_bi & (((SFS_LONG)-1) >> 8)) * ROOT_entry.bytes_used) + ((o_bi >> (8 * (sizeof(SFS_LONG) - 1))) * sizeof(SFS_ENTRY));
                    returnOk = SFSPort_Write(ptr, (U08 OS_HUGE *)&nEntry, sizeof(nEntry.name_hash));
                } else {
                    returnOk = SFS_NAME_EXIST;             // new-name exist
                }
            }
        } else {
            returnOk = SFS_PATH_ERR;                       // error in new-name
        }
    } else {
        returnOk = SFS_PATH_ERR;                           // no existing PATH given
    }
    SFS_User[user]->currentEntry = currDir;                // restore actual position
    SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                         CREATE FILE                      *
************************************************************
*/
U08 SFS_CreateFile(U08 OS_HUGE *Name, SFS_ATTR Attribs, SFS_LONG size) KC51_REENTRANT
{
    SFS_ENTRY   currDir;
    U08         returnOk, user;
    SFS_ATTR    a;
    SFS_LONG    Pos = 0;

    returnOk = SFSCheckUser(&user);
    if(returnOk != SFS_NO_ERR) {
        return(returnOk);
    }
    a = Attribs & SFS_ATTR_RW; 
    if(a == 0) {
        SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
        return(SFS_WRONG_A);
    }
    a |= SFS_ATTR_FILE;
    currDir = SFS_User[user]->currentEntry;                // save actual position and go into PATH
    returnOk = SFSChange_Entry(Name, 0);                   // go up to last part of path
    if(returnOk == SFS_NO_ERR) {                           // create Entry
        Pos = SFSstrrpos(Name, '/');
        if(Pos == (SFS_LONG)(0-1)) {
            Pos = 0;
        } else {
            Pos++;
        }
        if(Pos < SFSstrlen(Name)) {                        // create Entry
            returnOk = SFSCreate_Entry(&Name[Pos], user, a, size);
        } else {
            returnOk = SFS_PATH_ERR;
        }
    }
    SFS_User[user]->currentEntry = currDir;
    SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                         REMOVE FILE                      *
************************************************************
*/
U08 SFS_RemoveFile(U08 OS_HUGE *Name) KC51_REENTRANT
{
    SFS_ENTRY   currDir, cEntry;
    SFS_LONG    ptr;
    U08         returnOk, user;

    returnOk = SFSCheckUser(&user);
    if(returnOk != SFS_NO_ERR) {
        return(returnOk);
    }
    currDir = SFS_User[user]->currentEntry;                // save actual position and go into file
    returnOk = SFSChange_Entry(Name, 1);                   // go into file / into linked file
    if(returnOk == SFS_NO_ERR) {
        if((SFS_User[user]->currentEntry.attr__block_cnt & (SFS_ATTR_LINKS_M << (8 * (sizeof(SFS_LONG) - 1)))) == 0) {
            returnOk = SFSFileOpend(&SFS_User[user]->currentEntry);
            if(returnOk == SFS_NO_ERR) {
                returnOk = SFSMem_Free(SFS_User[user]->currentEntry.block_start, SFS_User[user]->currentEntry.attr__block_cnt & (((SFS_LONG)(-1)) >> 8));
                if(returnOk == SFS_NO_ERR) {
                    cEntry.name_hash = 0;
                    ptr = ((SFS_User[user]->h_entrybi & (((SFS_LONG)-1)) >> 8) * ROOT_entry.bytes_used) + ((SFS_User[user]->h_entrybi >> (8 * (sizeof(SFS_LONG) - 1))) * sizeof(SFS_ENTRY));
                    returnOk = SFSPort_Write(ptr, (U08 OS_HUGE *)&cEntry, sizeof(cEntry.name_hash));
                }
            }
        } else {
            returnOk = SFS_LINKED;                         // entry is linked
        }
    }
    SFS_User[user]->currentEntry = currDir;                // restore saved position
    SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                         CHANGE DIR                       *
************************************************************
*/
U08 SFS_ChangeDir(U08 OS_HUGE *Name) KC51_REENTRANT
{
    U08 returnOk, user;

    returnOk = SFSCheckUser(&user);
    if(returnOk != SFS_NO_ERR) {
        return(returnOk);
    }
    returnOk = SFSChange_Entry(Name, 2);                   // go into dir / into linked dir
    SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
    return(returnOk);
}

#if (SFS_TempDIR != 0)
/*
************************************************************
*                    CHANGE DIR TEMPORARY                  *
************************************************************
*/
U08 SFS_ChangeDirTemp(U08 OS_HUGE *Name) KC51_REENTRANT
{
    U08   returnOk, user;

    returnOk = SFSCheckUser(&user);
    if(returnOk != SFS_NO_ERR) {
        return(returnOk);
    }
    if(SFS_User[user]->TempDir.name_hash != 0) {
        SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
        return(SFS_TMP_DIR);
    }
    SFS_User[user]->TempDir = SFS_User[user]->currentEntry; // save current dir
    returnOk = SFSChange_Entry(Name, 2);                   // go into dir / into linked dir
    SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                 RESTORE DIR FROM TEMPORARY               *
************************************************************
*/
U08 SFS_BackDirTemp(void) KC51_REENTRANT
{
    U08                  returnOk, user;

    returnOk = SFSCheckUser(&user);
    if(returnOk != SFS_NO_ERR) {
        return(returnOk);
    }
    SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
    if(SFS_User[user]->TempDir.name_hash != 0) {           // restore saved current dir
        SFS_User[user]->currentEntry = SFS_User[user]->TempDir;
        SFS_User[user]->TempDir.name_hash = 0;
        return(SFS_NO_ERR);
    }
    return(SFS_TMP_DIR);
}
#endif

/*
************************************************************
*                    GET NAMED ENTRY-INFOS                 *
************************************************************
*/
U08 SFS_GetEntry(U08 OS_HUGE *Name, SFS_GET OS_HUGE *get) KC51_REENTRANT
{
    SFS_ENTRY   currDir, fEntry;
    U08         returnOk, user;

    returnOk = SFSCheckUser(&user);
    if(returnOk != SFS_NO_ERR) {
        return(returnOk);
    }
    currDir = SFS_User[user]->currentEntry;                // save actual position and go into file
    returnOk = SFSChange_Entry(Name, 4);                   // go into * / stay on link-*
    if(returnOk == SFS_NO_ERR) {
        fEntry = SFS_User[user]->currentEntry;
        get->name_hash = fEntry.name_hash;
        get->Attr      = fEntry.attr__block_cnt >> (8 * (sizeof(SFS_LONG) - 1)) & (SFS_ATTR_DIR | SFS_ATTR_FILE | SFS_ATTR_LINK | SFS_ATTR_WO | SFS_ATTR_RO);
        get->Length    = fEntry.bytes_used;
        if((fEntry.attr__block_cnt & (SFS_ATTR_LINKS_M << (8 * (sizeof(SFS_LONG) - 1)))) != 0) {
            get->Linked = TRUE;
        } else {
            get->Linked = FALSE;
        }
    }
    SFS_User[user]->currentEntry = currDir;                // restore saved position
    SFS_User[user]->ErrNo = SFSMutex_Put(&SFS_mutex);
    return(returnOk);
}

/*
************************************************************
*              CHECK IF FILE IS OPEND (internal)           *
************************************************************
*/
U08 SFSFileOpend(SFS_ENTRY OS_HUGE *EntryPTR) KC51_REENTRANT
{
    U08 i, j;

    for(i = 0; i < SFS_MAX_USER; i++) {
        if(SFS_User[i] != (SFS_USER OS_HUGE *)NULL) {
            for(j = 0; j < SFS_HANDLES; j++) {
                if(SFS_User[i]->used[j].openEntry.block_start == EntryPTR->block_start) {
                    return(SFS_FILE_OPEN);
                }
            }
        }
    }
    return(SFS_NO_ERR);
}

/*
************************************************************
*                   CHANGE ENTRY (internal)                *
*                                                          *
*   mode:   0 - go up to last part of path                 *
*           1 - go into file / into linked file            *
*           2 - go into dir  / into linked dir             *
*           3 - go into *    / into link-*                 *
*           4 - go into *    / stay on link-*              *
*           5 - go into dir  / stay on link-dir            *
*                                                          *
************************************************************
*/
U08 SFSChange_Entry(U08 OS_HUGE *Name, U08 mode) KC51_REENTRANT
{
    SFS_ENTRY            checkEntry;
    SFS_LONG             Pos = 0, Pos1, block_nr = 0, free_bi = 0, i, ptr;
    U08                  returnOk = SFS_NO_ERR, user;

    user = SFSGetUserByID();
    if(user == 0xFF) {
        return(SFS_NO_USER);
    }
    checkEntry = SFS_User[user]->currentEntry;
    if(Name[Pos] == '/') {                                 // is from ROOT (for all root-symbols at front)
        checkEntry = ROOT_entry;
        while(Name[Pos] == '/') Pos++;
        if(Pos == SFSstrlen(Name)) {                       // end of path after only root...
            SFS_User[user]->currentEntry = checkEntry;     // save entry and block-nr 0
            SFS_User[user]->h_entrybi    = 0;
            return(returnOk);
        }
    }
    if((Name[Pos] == '.') && (Name[Pos + 1] == '/')) {
        Pos += 2;
        while(Name[Pos] == '/') Pos++;
    }
    if(0 != SFS_hash_compare((U08 OS_HUGE *)"/", 1, checkEntry.name_hash))
        block_nr = checkEntry.block_start;
    while(Pos < SFSstrlen(Name)) {                         // up to end of PATH
        do {                                               // search end of next part of PATH
            Pos1 = SFSstrpos(&Name[Pos],'/');
            if(Pos1 == 0) Pos++;
        } while(Pos1 == 0);
        if(Pos1 == (SFS_LONG)(0-1)) {                      // if last part of PATH, calc length
            Pos1 = SFSstrlen(Name)-Pos;
        }
        i = 0;
        while(i < ROOT_entry.attr__block_cnt) {
            returnOk = SFSPort_Read((block_nr * ROOT_entry.bytes_used) + (i * sizeof(SFS_ENTRY)), (U08 OS_HUGE *)&checkEntry, sizeof(SFS_ENTRY));
            if(returnOk == SFS_NO_ERR) {
                if(0 == SFS_hash_compare(&Name[Pos], Pos1, checkEntry.name_hash)) break;
            } else {
                return(returnOk);
            }
            if((i != 0) && (free_bi == 0) && ((Pos + Pos1) == SFSstrlen(Name)) && (mode == 0) && (checkEntry.name_hash == 0)) {
                free_bi = (block_nr & (((SFS_LONG)-1) >> 8)) | (i << (8 * (sizeof(SFS_LONG) - 1)));
            }
            i++;
        }
        if(i < ROOT_entry.attr__block_cnt) {
            Pos += Pos1; Pos1 = 0;                         // found ...
            if(Pos < SFSstrlen(Name)) {                    // ... next level if DIR
                if(((checkEntry.attr__block_cnt >> (8 * (sizeof(SFS_LONG) - 1))) & ~(SFS_ATTR_LINKS_M | SFS_ATTR_LINK)) == SFS_ATTR_DIR) {
                    while(((checkEntry.attr__block_cnt >> (8 * (sizeof(SFS_LONG) - 1))) & SFS_ATTR_LINK)) {
                        ptr = ((checkEntry.block_start & (((SFS_LONG)-1)) >> 8) * ROOT_entry.bytes_used) + ((checkEntry.block_start >> (8 * (sizeof(SFS_LONG) - 1))) * sizeof(SFS_ENTRY));
                        returnOk = SFSPort_Read(ptr, (U08 OS_HUGE *)&checkEntry, sizeof(SFS_ENTRY));
                        if(returnOk != SFS_NO_ERR) {
                            return(returnOk);
                        }
                        if(((checkEntry.attr__block_cnt >> (8 * (sizeof(SFS_LONG) - 1))) & ~(SFS_ATTR_LINKS_M | SFS_ATTR_LINK)) != SFS_ATTR_DIR) {
                            return(SFS_PATH_ERR);          // link is no DIR -> error in PATH
                        }
                        if(((checkEntry.attr__block_cnt >> (8 * (sizeof(SFS_LONG) - 1))) & SFS_ATTR_LINKS_M) == 0) {
                            return(SFS_NO_LINK);           // entry itself is not linked
                        }
                    }
                    block_nr = checkEntry.block_start;
                    continue;
                } else {
                    return(SFS_PATH_ERR);                  // no DIR -> error in PATH
                }
            } else {                                       // last part of PATH entry found ...
                if(mode == 0) {                            // ... if  mode-0
                    return(SFS_NAME_EXIST);
                }
                if(mode < 4) {                             // if not to stay on link-entry itself
                    if((mode == 3) ||
                      ((mode == 1) &&  ((checkEntry.attr__block_cnt >> (8 * (sizeof(SFS_LONG) - 1))) & SFS_ATTR_FILE)) ||
                      ((mode == 2) && !((checkEntry.attr__block_cnt >> (8 * (sizeof(SFS_LONG) - 1))) & SFS_ATTR_FILE))) {
                        while(((checkEntry.attr__block_cnt >> (8 * (sizeof(SFS_LONG) - 1))) & SFS_ATTR_LINK)) {
                            ptr = ((checkEntry.block_start & (((SFS_LONG)-1)) >> 8) * ROOT_entry.bytes_used) + ((checkEntry.block_start >> (8 * (sizeof(SFS_LONG) - 1))) * sizeof(SFS_ENTRY));
                            returnOk = SFSPort_Read(ptr, (U08 OS_HUGE *)&checkEntry, sizeof(SFS_ENTRY));
                            if(returnOk != SFS_NO_ERR) {
                                return(returnOk);
                            }
                            if(((checkEntry.attr__block_cnt >> (8 * (sizeof(SFS_LONG) - 1))) & SFS_ATTR_LINKS_M) == 0) {
                                return(SFS_NO_LINK);       // entry itself is not linked
                            }
                        }
                    } else {
                        return(SFS_PATH_ERR);              // last part of PATH is wrong
                    }
                }
                SFS_User[user]->currentEntry = checkEntry; // save entry and block-nr/offs
                SFS_User[user]->h_entrybi    = (block_nr & (((SFS_LONG)-1) >> 8)) | (i << (8 * (sizeof(SFS_LONG) - 1)));
            }
            break;
        }                                                  // not found ...
        if(((Pos + Pos1) == SFSstrlen(Name)) && (mode == 0)) {
            Pos = SFSstrlen(Name); Pos1 = 0;               // ... if last part of PATH and mode-0
            if(0 != free_bi) {
                SFS_User[user]->h_entrybi = free_bi;
                returnOk = SFS_NO_ERR;
            } else {
                returnOk = SFS_NO_ENTRY;
            }
        } else {
            if((Pos + Pos1) < SFSstrlen(Name)) {
                returnOk = SFS_PATH_ERR;                   // wrong element in PATH
            } else {
                returnOk = SFS_NOT_EXIST;                  // last part of PATH not exist
            }
            Pos = SFSstrlen(Name); Pos1 = 0;
        }
    }
    return(returnOk);
}

/*
************************************************************
*     CREATE DIR/FILE ENTRY incl. Data Block (internal)    *
************************************************************
*/
U08 SFSCreate_Entry(U08 OS_HUGE *Name, U08 user, SFS_ATTR Attr, SFS_LONG size_b) KC51_REENTRANT
{
    SFS_ENTRY   newEntry;
    SFS_LONG    block_s, a_cnt = 0, b;
    U08         returnOk;

    b = (size_b / ROOT_entry.bytes_used) + ((size_b % ROOT_entry.bytes_used)? 1 : 0);
    returnOk = SFSMem_Alloc(b, &block_s);
    if(returnOk == SFS_NO_ERR) {
        returnOk = SFSPort_Fill(block_s * ROOT_entry.bytes_used, 0, size_b);
        if(returnOk == SFS_NO_ERR) {
            switch(Attr & SFS_ATTR_FILE) {
                case SFS_ATTR_DIR:
                    newEntry.name_hash       = SFS_hash_calc((U08 OS_HUGE *)"..", 2);
                    newEntry.attr__block_cnt = 0;
                    newEntry.block_start     = (SFS_User[user]->h_entrybi & (((SFS_LONG)-1) >> 8));
                    newEntry.bytes_used      = 0;
                    returnOk = SFSPort_Write((block_s * ROOT_entry.bytes_used), (U08 OS_HUGE *)&newEntry, sizeof(SFS_ENTRY));
                    if (SFS_NO_ERR != returnOk) return(returnOk);
                    a_cnt = ((SFS_LONG)SFS_ATTR_DIR << (8 * (sizeof(SFS_LONG) - 1)));
                    break;
                case SFS_ATTR_FILE:
                    a_cnt = ((SFS_LONG)Attr << (8 * (sizeof(SFS_LONG) - 1))) | b;
                    break;
                default:
                    break;
            }
            newEntry.name_hash       = SFS_hash_calc(Name, SFSstrlen(Name));
            newEntry.attr__block_cnt = a_cnt;
            newEntry.block_start     = block_s;
            newEntry.bytes_used      = 0;
            b = ((SFS_User[user]->h_entrybi & (((SFS_LONG)-1) >> 8)) * ROOT_entry.bytes_used) + ((SFS_User[user]->h_entrybi >> (8 * (sizeof(SFS_LONG) - 1))) * sizeof(SFS_ENTRY));
            returnOk = SFSPort_Write(b, (U08 OS_HUGE *)&newEntry, sizeof(SFS_ENTRY));
        }
    }
    return(returnOk);
}

/*
************************************************************
*                 FORMAT SFS-MEMORY (internal)             *
************************************************************
*/
static U08 BLOCK_map[SFS_BLOCKMAP_SIZE];                   // blocks-bitmap (1 = free / 0 = used)

U08 SFSMem_Format(void) KC51_REENTRANT
{
    U08      returnOk, n;
    SFS_LONG i = 0, cnt;

    cnt = ((ROOT_entry.attr__block_cnt * sizeof(SFS_ENTRY)) / ROOT_entry.bytes_used) + (((ROOT_entry.attr__block_cnt * sizeof(SFS_ENTRY)) % ROOT_entry.bytes_used)? 1 : 0);
    while(cnt > 0) {
        n = (cnt > 8)? 8 : cnt;
        BLOCK_map[i] = ((U08)-1) >> n;
        i++; cnt -= n;
    }
    for (; i < sizeof(BLOCK_map); i++) {
        BLOCK_map[i] = (U08)-1;
    }
    returnOk = SFSPort_Write(ROOT_entry.block_start * ROOT_entry.bytes_used - sizeof(BLOCK_map), (U08 OS_HUGE *)&BLOCK_map, sizeof(BLOCK_map));
    return(returnOk);
}

/*
************************************************************
*                 FORMAT SFS-MEMORY (internal)             *
************************************************************
*/
U08 SFSMem_Check(void) KC51_REENTRANT
{
    U08      returnOk, m;

    returnOk = SFSPort_Read(ROOT_entry.block_start * ROOT_entry.bytes_used - (ROOT_entry.block_start / 8), (U08 OS_HUGE *)&BLOCK_map, (ROOT_entry.block_start / 8));
    if(returnOk == SFS_NO_ERR) {                           // check BLOCK_map at the end of memory
        m = (U08)-1; m = m >> ROOT_entry.attr__block_cnt;
        if((BLOCK_map[0] & ~m) != 0) {
            return(SFS_FORMAT_ERR);
        }
    }
    return(SFS_NO_ERR);
}

/*
************************************************************
*           SEARCH & ALLOCATE SFS-MEMORY (internal)        *
************************************************************
*/
U08 SFSMem_Alloc(SFS_LONG blocks, SFS_LONG *block_start) KC51_REENTRANT
{
    SFS_LONG block_s, i, j, cnt;
    U08      returnOk, m;

    returnOk = SFSPort_Read(ROOT_entry.block_start * ROOT_entry.bytes_used - (ROOT_entry.block_start / 8), (U08 OS_HUGE *)&BLOCK_map, (ROOT_entry.block_start / 8));
    if(returnOk == SFS_NO_ERR) {                           // scan & allocate n-blocks in BLOCK_map and return block_start of found location
        cnt = 0; block_s = 0;
        for(i = 0; i < (ROOT_entry.block_start / 8); i++) {
            m = 0x80; j = 0;
            while(m != 0) {
                if((BLOCK_map[i] & m) != 0) {
                    if(block_s == 0) block_s = i * 8 + j;
                    cnt++;
                    if(cnt == blocks) {                    // sufficient continues blocks found, allocate them
                        *block_start = block_s;
                        if(((block_s % 8) + blocks) <= 8) {
                            BLOCK_map[block_s /8] &= ~((U08)(0xFF << (8 - blocks)) >> (block_s % 8));
                        } else {
                            BLOCK_map[block_s /8] &= ~(0xFF >> (block_s % 8));
                            blocks -= (8 - (block_s % 8)); block_s += (8 - (block_s % 8));
                            while(blocks >= 8) {
                                BLOCK_map[block_s /8] = 0x00; 
                                block_s += 8; blocks -= 8;
                            }
                            if(blocks > 0) {
                                BLOCK_map[block_s /8] &= ~(0xFF << (8 - blocks));
                            }
                        }
                        returnOk = SFSPort_Write(ROOT_entry.block_start * ROOT_entry.bytes_used - (ROOT_entry.block_start / 8), (U08 OS_HUGE *)&BLOCK_map, (ROOT_entry.block_start / 8));
                        return(returnOk);
                    }
                } else {
                    cnt = 0; block_s = 0;
                }
                j++; m >>= 1;
            }
        }
    }
    *block_start = 0;
    return(SFS_MEM_OVF);
}

/*
************************************************************
*                  FREE SFS-MEMORY (internal)              *
************************************************************
*/
U08 SFSMem_Free(SFS_LONG block_start, SFS_LONG blocks) KC51_REENTRANT
{
    U08 returnOk;

    returnOk = SFSPort_Read(ROOT_entry.block_start * ROOT_entry.bytes_used - (ROOT_entry.block_start / 8), (U08 OS_HUGE *)&BLOCK_map, (ROOT_entry.block_start / 8));
    if(returnOk == SFS_NO_ERR) {                           // free n-blocks from block_start in BLOCK_map
        if(((block_start % 8) + blocks) <= 8) {
            BLOCK_map[block_start /8] |= ((U08)(0xFF << (8 - blocks)) >> (block_start % 8));
        } else {
            BLOCK_map[block_start /8] |= (0xFF >> (block_start % 8));
            blocks -= (8 - (block_start % 8)); block_start += (8 - (block_start % 8));
            while(blocks >= 8) {
                BLOCK_map[block_start /8] = 0xFF; 
                block_start += 8; blocks -= 8;
            }
            if(blocks > 0) {
                BLOCK_map[block_start /8] |= (0xFF << (8 - blocks));
            }
        }
        returnOk = SFSPort_Write(ROOT_entry.block_start * ROOT_entry.bytes_used - (ROOT_entry.block_start / 8), (U08 OS_HUGE *)&BLOCK_map, (ROOT_entry.block_start / 8));
    }
    return(returnOk);
}

#if (SFS_REMOVEDIRTREE != 0)
 #include "SFS_RemoveDirTree.c"
#endif

/************************* END ****************************/
